/*--------------------------------------------------------------- 
 * Copyright (c) 1999,2000,2001,2002                              
 * The Board of Trustees of the University of Illinois            
 * All Rights Reserved.                                           
 *--------------------------------------------------------------- 
 * Permission is hereby granted, free of charge, to any person    
 * obtaining a copy of this software (Iperf) and associated       
 * documentation files (the "Software"), to deal in the Software  
 * without restriction, including without limitation the          
 * rights to use, copy, modify, merge, publish, distribute,        
 * sublicense, and/or sell copies of the Software, and to permit     
 * persons to whom the Software is furnished to do
 * so, subject to the following conditions: 
 *
 *     
 * Redistributions of source code must retain the above 
 * copyright notice, this list of conditions and 
 * the following disclaimers. 
 *
 *     
 * Redistributions in binary form must reproduce the above 
 * copyright notice, this list of conditions and the following 
 * disclaimers in the documentation and/or other materials 
 * provided with the distribution. 
 * 
 *     
 * Neither the names of the University of Illinois, NCSA, 
 * nor the names of its contributors may be used to endorse 
 * or promote products derived from this Software without
 * specific prior written permission. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL THE CONTIBUTORS OR COPYRIGHT 
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
 * ________________________________________________________________
 * National Laboratory for Applied Network Research 
 * National Center for Supercomputing Applications 
 * University of Illinois at Urbana-Champaign 
 * http://www.ncsa.uiuc.edu
 * ________________________________________________________________ 
 *
 * PerfSocket.cpp 
 * by Mark Gates <mgates@nlanr.net> 
 * &  Ajay Tirumala <tirumala@ncsa.uiuc.edu> 
 * ------------------------------------------------------------------- 
 * Has routines the Client and Server classes use in common for 
 * performance testing the network. 
 * ------------------------------------------------------------------- 
 * headers 
 * uses 
 *   <stdlib.h> 
 *   <stdio.h> 
 *   <string.h> 
 * 
 *   <sys/types.h> 
 *   <sys/socket.h> 
 *   <unistd.h> 
 * 
 *   <arpa/inet.h> 
 *   <netdb.h> 
 *   <netinet/in.h> 
 *   <sys/socket.h> 
 * ------------------------------------------------------------------- */ 

#define HEADERS() 
 
#include "headers.h" 
 
#include "PerfSocket.hpp" 
#include "Settings.hpp" 
#include "Locale.hpp" 
#include "delay.hpp" 
#include "Listener.hpp" 
#include "util.h" 
 
const double kSecs_to_usecs = 1e6; 
const int    kBytes_to_Bits = 8; 
 
using namespace std; 
 
/* ------------------------------------------------------------------- 
 * Send data using the connected UDP socket, 
 * until a termination flag is received. 
 * Does not close the socket. 
 * ------------------------------------------------------------------- */ 
 
void PerfSocket::Send_UDP( void ) 
{ 
  long currLen; 
  int32_t datagramID = 0; 
  struct UDP_datagram* mBuf_UDP = (struct UDP_datagram*) mBuf; 
 
  int delay_target; 
  int delay = 0; 
  int adjust; 
 
  // terminate loop nicely on user interupts 
  sInterupted = false; 
  my_signal( SIGINT, Sig_Interupt ); 
 
  // compute delay for UDP bandwidth restriction, constrained to [0,1] seconds 
  delay_target = (int) (mBufLen * ((kSecs_to_usecs * kBytes_to_Bits) 
                                     / gSettings->GetUDPRate())); 
  if ( delay_target < 0  || 
       delay_target > (int) 1 * kSecs_to_usecs ) { 
    printf( warn_delay_large, delay_target / kSecs_to_usecs ); 
    delay_target = (int) kSecs_to_usecs * 1; 
  } 
 
  // check if we're sending multicast, and set TTL 
  // TODO is this the right place to put this??? 
  SocketAddr remote = getRemoteAddress(); 
  if ( remote.isMulticast()) { 
    McastSetTTL( gSettings->GetMcastTTL() ); 
    connect(mSock,(struct sockaddr *)&remote,(int)sizeof(remote)); 
  } 
 
  Timestamp lastPacketTime; 
  InitTransfer(); 
  // File Input option 
  bool fileInput = gSettings->GetFileInput(); 
   
  // Indicates if the stream is readable 
  bool canRead; 
 
  // Due to the UDP timestamps etc, included 
  // reduce the read size by an amount 
  // equal to the header size 
  if (fileInput) 
    extractor->reduceReadSize(sizeof(struct UDP_datagram)); 
 
  mStartTime.setnow(); 
 
  do { 
    mPacketTime.setnow(); 
 
    // Test case: drop 17 packets and send 2 out-of-order: 
    // sequence 51, 52, 70, 53, 54, 71, 72 
    //switch( datagramID ) { 
    //  case 53: datagramID = 70; break; 
    //  case 71: datagramID = 53; break; 
    //  case 55: datagramID = 71; break; 
    //  default: break; 
    //} 
 
    // store datagram ID into buffer 
    mBuf_UDP->id      = htonl( datagramID++ ); 
    mBuf_UDP->tv_sec  = htonl( mPacketTime.getSecs()); 
    mBuf_UDP->tv_usec = htonl( mPacketTime.getUsecs()); 
     
    // Read the next data block from 
    // the file if it's file input 
    if (fileInput) { 
      extractor->getNextDataBlock(mBuf + sizeof(struct UDP_datagram)); 
      canRead = extractor->canRead(); 
    } 
    else 
      canRead = true; 
     
    // perform write 
    currLen = write( mSock, mBuf, mBufLen ); 
    if ( currLen < 0 ) { 
      WARN_errno( currLen < 0, "write" ); 
      break; 
    } 
    mTotalLen += currLen; 
     
    // delay between writes 
    // make an adjustment for how long the last loop iteration took 
    // TODO this doesn't work well in certain cases, like 2 parallel streams 
    adjust = delay_target - mPacketTime.subUsec( lastPacketTime ); 
    lastPacketTime = mPacketTime; 
 
    if ( adjust > 0  ||  delay > 0 ) { 
      delay += adjust; 
    } 
    if ( delay > 0 ) { 
      delay_loop( delay ); 
    } 
 
    // periodically report bandwidths 
    ReportPeriodicBW(); 
  } while( ! (sInterupted  || 
              (mMode_time   &&  mPacketTime.after( mEndTime ))  || 
              (!mMode_time  &&  mTotalLen >= mAmount)) && canRead); 
               
  // stop timing 
  mEndTime.setnow(); 
  ReportBW( mTotalLen, 0.0, mEndTime.subSec( mStartTime )); 
 
  // send a final terminating datagram 
  // Don't count in the mTotalLen. The server counts this one, 
  // but didn't count our first datagram, so we're even now. 
  // The negative datagram ID signifies termination to the server. 
  mPacketTime.setnow(); 
 
  // store datagram ID into buffer 
  mBuf_UDP->id      = htonl( -datagramID  ); 
  mBuf_UDP->tv_sec  = htonl( mPacketTime.getSecs()); 
  mBuf_UDP->tv_usec = htonl( mPacketTime.getUsecs()); 
 
  if ( remote.isMulticast()) { 
    write( mSock, mBuf, mBufLen ); 
  } 
  else { 
    write_UDP_FIN( mSock, mBuf, mBufLen ); 
  } 
 
  printf( report_datagrams, mSock, datagramID ); 
} 
// end SendUDP 
 
/* ------------------------------------------------------------------- 
 * Receieve data from the (connected) UDP socket. 
 * Sends termination flag several times at the end. 
 * Does not close the socket. 
 * ------------------------------------------------------------------- */ 
 
void PerfSocket::Recv_UDP( void ) 
{ 
  bool going = true; 
  long currLen; 
  int32_t datagramID     = 0; 
  int32_t lastDatagramID = 0; 
  struct UDP_datagram* mBuf_UDP  = (struct UDP_datagram*) mBuf; 
  int errorCnt   = 0; 
  int outofOrder = 0; 
 
  extern Mutex clients_mutex; 
 
  // for jitter 
  Timestamp sentTime; 
  double transit; 
  double lastTransit = 0.0; 
  double deltaTransit; 
 
  InitTransfer(); 
  do { 
    // perform read 
    currLen = read( mSock, mBuf, mBufLen ); 
     
    mPacketTime.setnow(); 
    mTotalLen += currLen; 
 
    // read the datagram ID and sentTime out of the buffer 
    datagramID = ntohl( mBuf_UDP->id ); 
    sentTime.set( ntohl( mBuf_UDP->tv_sec  ), 
                  ntohl( mBuf_UDP->tv_usec )); 
 
    // TODO functionalize this 
    // from RFC 1889, Real Time Protocol (RTP) 
    // J = J + ( | D(i-1,i) | - J ) / 16 
    transit = mPacketTime.subSec( sentTime ); 
    if ( lastTransit != 0.0 ) { 
      deltaTransit = transit - lastTransit; 
      if ( deltaTransit < 0.0 ) { 
        deltaTransit = -deltaTransit; 
      } 
      mJitter += (1.0/16.0) * (deltaTransit - mJitter); 
    } 
    lastTransit = transit; 
 
    // terminate when datagram begins with negative index 
    // the datagram ID should be correct, just negated 
    if ( datagramID < 0 ) { 
      going = false; 
      datagramID = -datagramID; 
    } 
    // packet loss occured if the datagram numbers aren't sequential 
    if ( datagramID != lastDatagramID+1 ) { 
      if ( datagramID < lastDatagramID+1 ) { 
        outofOrder++; 
      } 
      else { 
        errorCnt += datagramID - (lastDatagramID+1); 
      } 
    } 
    // never decrease datagramID (e.g. if we get an out-of-order packet) 
    if ( datagramID > lastDatagramID ) { 
      lastDatagramID = datagramID; 
    } 
 
    // periodically report bandwidths 
    ReportPeriodicBW_Jitter_Loss( errorCnt, outofOrder, datagramID ); 
  } while( going ); 
 
  // stop timing 
  mEndTime.setnow(); 
  ReportBW_Jitter_Loss( mTotalLen, 0.0, mEndTime.subSec( mStartTime ), 
                        errorCnt, outofOrder, datagramID ); 
 
  // send a acknowledgement back only if we're NOT receiving multicast 
  SocketAddr local = getLocalAddress(); 
  if ( ! local.isMulticast()) { 
    // send back an acknowledgement of the terminating datagram 
    write_UDP_AckFIN( mSock, mBuf, mBufLen ); 
  } 
  else { 
    // get the remote address and remove it from the set of clients 
    (clients_mutex).Lock();     
    SocketAddr remote = getRemoteAddress(); 
#if defined(IPV6)
    sockaddr_storage peer = *(sockaddr_storage *) (remote.get_sockaddr()); 
#else
    sockaddr peer = *(sockaddr *) (remote.get_sockaddr()); 
#endif
    Multicast_remove_client(peer); 
    (clients_mutex).Unlock(); 
  } 
} 
// end RecvUDP 
 
/* ------------------------------------------------------------------- 
 * Do the equivalent of an accept() call for UDP sockets. This waits 
 * on a listening UDP socket until we get a datagram. Connect the 
 * UDP socket for efficiency. 
 * Verify that the datagram ID starts at zero. TODO is that needed still? 
 * ------------------------------------------------------------------- */ 

  // 10feb02(mjz): convert sockaddr_in to void in returned type
  //               convert to using sockaddr_storage internally 
void PerfSocket::Accept_UDP( void ) 
{ 
  int32_t datagramID; 
  struct UDP_datagram* mBuf_UDP  = (struct UDP_datagram*) mBuf; 
  //  struct sockaddr_in peer;  MJZ -- IPv4 only gotta change...
  // try to do "peer" in proto-independant way.  in FreeBSD 4.3-rel (KAME)
  // they suggest in socket.h using sockaddr_storage struct.  max is 128.
  // above, for sockaddr, it mentions 255 being longest address structure.
  // so, for now was going to fudge completely... but maybe that's bad.
  // use sockaddr_storage.  not sure if ports.
#if defined(IPV6)
  struct sockaddr_storage peer_storage;
#else
  struct sockaddr peer_storage;
#endif
  struct sockaddr *peer = (sockaddr *) &peer_storage;
  Socklen_t peerlen; 
  int rc; 
 
  // TODO push some of this down into the Socket class 
  // for UDP read the first datagram before the timer since 
  // we generally have to block a long time before this one. 
  // we also get the client's address and connect to it. 
  do { 
    //peerlen = sizeof(peer);  MJZ
    peerlen = sizeof(peer_storage);
    rc = recvfrom( mSock, mBuf, mBufLen, 0, 
                   peer, &peerlen ); 
    FAIL_errno( rc == SOCKET_ERROR, "recvfrom" ); 
  
    datagramID = ntohl( mBuf_UDP->id ); 
  } while( datagramID < 0 ); 

#ifdef DBG_MJZ 
  {
  // DBG MJZ
  // note, return from recvfrom is sockaddr_in???
    struct sockaddr * sa = (sockaddr *) peer;
  fprintf(stderr, "RF len: %d salen: %d fam: %d data chars %2x %2x %2x %2x ...\n"
,  
        peerlen, sa->sa_len, sa->sa_family, sa->sa_data[0], 
	sa->sa_data[1], sa->sa_data[2], sa->sa_data[3]);
  }
#endif /* DBG_MJZ */
 
  rc = connect( mSock, peer, peerlen ); 
  FAIL_errno( rc == SOCKET_ERROR, "connect" ); 
  return; 
 
} 
 
/* ------------------------------------------------------------------- 
 * Send a datagram on the socket. The datagram's contents should signify 
 * a FIN to the application. Keep re-transmitting until an 
 * acknowledgement datagram is received. 
 * ------------------------------------------------------------------- */ 
 
void PerfSocket::write_UDP_FIN( int sock, void* buf, int len ) 
{ 
  int rc; 
  fd_set readSet; 
  struct timeval timeout; 
 
  int count = 0; 
  while( count < 10 ) { 
    count++; 
 
    // write data 
    write( sock, buf, len ); 
 
    // wait until the socket is readable, or our timeout expires 
    FD_ZERO( &readSet ); 
    FD_SET( sock, &readSet ); 
    timeout.tv_sec  = 0; 
    timeout.tv_usec = 250000; // quarter second, 250 ms 
 
    rc = select( sock+1, &readSet, NULL, NULL, &timeout ); 
    FAIL_errno( rc == SOCKET_ERROR, "select" ); 
 
    if ( rc == 0 ) { 
      // select timed out 
      continue; 
    } 
    else { 
      // socket ready to read 
      rc = read( sock, buf, len ); 
      FAIL_errno( rc < 0, "read" ); 
      return; 
    } 
  } 
 
  printf( warn_no_ack, mSock, count ); 
} 
// end write_UDP_FIN 
 
/* ------------------------------------------------------------------- 
 * Send an AckFIN (a datagram acknowledging a FIN) on the socket, 
 * then select on the socket for some time. If additional datagrams 
 * come in, probably our AckFIN was lost and they are re-transmitted 
 * termination datagrams, so re-transmit our AckFIN. 
 * ------------------------------------------------------------------- */ 
 
void PerfSocket::write_UDP_AckFIN( int sock, void* buf, int len ) 
{ 
  int rc; 
 
  fd_set readSet; 
  FD_ZERO( &readSet ); 
 
  struct timeval timeout; 
 
  int count = 0; 
  while( count < 10 ) { 
    count++; 
 
    // write data 
    write( sock, buf, len ); 
 
    // wait until the socket is readable, or our timeout expires 
    FD_SET( sock, &readSet ); 
    timeout.tv_sec  = 1; 
    timeout.tv_usec = 0; 
 
    rc = select( sock+1, &readSet, NULL, NULL, &timeout ); 
    FAIL_errno( rc == SOCKET_ERROR, "select" ); 
 
    if ( rc == 0 ) { 
      // select timed out 
      return; 
    } 
    else { 
      // socket ready to read 
      rc = read( sock, buf, len ); 
      FAIL_errno( rc < 0, "read" ); 
      continue; 
    } 
  } 
 
  printf( warn_ack_failed, mSock, count ); 
} 
// end write_UDP_AckFIN 
 
#if defined(IPV6) 
void PerfSocket::Multicast_remove_client(sockaddr_storage peer) { 
  extern vector<sockaddr_storage> clients; 
#else
void PerfSocket::Multicast_remove_client(sockaddr peer) {
  extern vector<sockaddr> clients; 
#endif 
  for (int i=0; i < (int)(clients).size(); i++) { 
    struct sockaddr *sap = (struct sockaddr *)&peer, *sac = (struct sockaddr *) &clients[i];
#if defined(IPV6)
    sockaddr_storage t1 = (sockaddr_storage) clients[i];
#else
    sockaddr t1 = (sockaddr) clients[i];
#endif

#if defined(IPV6)

    if (sap->sa_family == AF_INET6 && sac->sa_family == AF_INET6) {
      // compare IPv6 addresses
      // hmm, what's the dumbest way to do this?
      
      sockaddr_in6* tp2 = (sockaddr_in6*) &t1;
      sockaddr_in6* peerp = (sockaddr_in6*) &peer;
      if (!memcmp(tp2->sin6_addr.s6_addr, peerp->sin6_addr.s6_addr, 16) &&
	  (tp2->sin6_port == peerp->sin6_port)) {
	
	(clients).erase(&(clients[i]));  
	return;
      }  
    }
#endif

    if (sap->sa_family == AF_INET && sac->sa_family == AF_INET) {
      // compare IPv4 adresses
      sockaddr_in* tp2 = (sockaddr_in*) &t1;
      sockaddr_in* peerp = (sockaddr_in*) &peer;
      if ( ((long) tp2->sin_addr.s_addr == (long) peerp->sin_addr.s_addr)
	   && ( tp2->sin_port == peerp->sin_port)) {
	
	(clients).erase(&(clients[i]));                                                   
	return;
      }
    }
  } 
  return; 
}
 
